from __future__ import annotations

from pathlib import Path
from typing import Callable, List, Union
from more_itertools import pairwise
import multiprocessing as mp
import numpy as np
import pandas as pd
from scipy.spatial import cKDTree
from ..logger import logger
from ..structs import (
    BoundingBox, TimePosition,
    UNIX_TIMESTAMP
)
from . import split
from ..decode.filedescriptor import (
    BaseColumns, Msg12318Columns, Msg5Columns
)
from .targetship import (
    TargetShip, AISMessage, 
    InterpolationError, Track, Targets
)
from ..utils import DataLoader

def _identity(x):
    return x


class NeighborhoodTreeSearch:
    """
    A class for performing neighborhood searches 
    on AIS message data using a kd-tree.

    This class takes AIS messages, filters them 
    based on a time window, and uses a kd-tree
    for efficient spatial searches to find neighboring 
    ships within a specified search radius.

    Attributes:
        `data` (DataLoader): Instance of DataLoader 
                             to load AIS message data.
        `time_delta` (int): Time delta in minutes for 
                            filtering messages around a 
                            specific time.
        `max_tgt_ships` (int): Maximum number of target 
                               ships to return in the 
                               neighborhood search.

    Methods:
        `get_neighbors`: Retrieves neighboring AIS messages 
                         within a specified search radius.
    """
    def __init__(self, 
                 data_loader: DataLoader,
                 time_delta: int,
                 max_tgt_ships: int) -> None:
        
        self.data = data_loader # Load data into memory
        self.time_delta = time_delta
        self.max_tgt_ships = max_tgt_ships
    
    def _build_kd_tree(self, data: pd.DataFrame) -> cKDTree:
        """
        Private method to build a kd-tree from latitude 
        and longitude columns of a pandas DataFrame.

        Args:
            `data` (pd.DataFrame): DataFrame containing 
                                   latitude and longitude data.

        Returns:
            `cKDTree`: A scipy cKDTree object built from 
                       the latitude and longitude data.

        Raises:
            AssertionError: If the input DataFrame lacks 
                            latitude or longitude columns.
        """
        assert Msg12318Columns.LAT.value in data and Msg12318Columns.LON.value in data, \
            "Input dataframe has no `lat` or `lon` columns"
        lat=data[Msg12318Columns.LAT.value].values
        lon=data[Msg12318Columns.LON.value].values
        return cKDTree(np.column_stack((lat,lon)))
            

    def get_neighbors(self, 
                      tpos: TimePosition, 
                      search_radius: float = 20) -> pd.DataFrame:
        """
        Return all AIS messages that are no more than
        `self.search_radius` [nm] away from the given position.

        Args:
            tpos: TimePosition object of postion and time for which 
                    neighbors shall be found        

        """
        if not self.data.loaded:
            self.data.load() # Load data into memory
        
        filtered = self._time_filter(
            self.data.dynamic_data,tpos.timestamp,self.time_delta
        )
        # Check if filterd result is empty
        if filtered.empty:
            logger.warning("No AIS messages found in time-filtered cell.")
            return filtered
        tree = self._build_kd_tree(filtered)
        # The conversion to degrees is only accurate at the equator.
        # Everywhere else, the distances get smaller as lines of 
        # Longitude are not parallel. Therefore, 
        # this is a conservative estimate.
        if not search_radius == np.inf:
            sr =  search_radius/60 # Convert to degrees
        else:
            sr = np.inf
        d, indices = tree.query(
            tpos.position.as_list,
            k=self.max_tgt_ships,
            distance_upper_bound=sr
        )
        # Sort out any infinite distances
        res = [indices[i] for i,j in enumerate(d) if j != float("inf")]

        return filtered.iloc[res]
    

    def _time_filter(self, 
                     df: pd.DataFrame, 
                     date: UNIX_TIMESTAMP, 
                     delta: int) -> pd.DataFrame:
        """
        Filter a pandas dataframe to only return 
        rows whose `Timestamp` is not more than 
        `delta` minutes apart from imput `date`.
        """
        assert BaseColumns.TIMESTAMP.value in df, "No `timestamp` column found"
        timezone = df[BaseColumns.TIMESTAMP.value].iloc[0].tz.zone # Get timezone
        date = pd.to_datetime(int(date),unit="s").tz_localize(timezone)
        dt = pd.Timedelta(delta, unit="minutes")
        mask = (
            (df[BaseColumns.TIMESTAMP.value] > (date-dt)) & 
            (df[BaseColumns.TIMESTAMP.value] < (date+dt))
        )
        return df.loc[mask]
class SearchAgent:
    """
    Class searching for target ships
    near a user-provided location.

    To initialize the Agent, you have to specify a 
    datapath, where source AIS Messages are saved as csv files.
    You also have to specify a global frame 
    for the search agent. Outside its borders, no
    search will be commenced.
    """
    
    def __init__(self, 
                 dynamic_paths: Union[Path,List[Path]],
                 frame: BoundingBox,
                 static_paths: Union[Path,List[Path]],
                 max_tgt_ships: int = 200,
                 preprocessor: Callable[[pd.DataFrame],pd.DataFrame] = _identity,
                 decoded: bool = True,
                 high_accuracy: bool = False) -> None:
       
        """ 
        frame: BoundingBox object setting the search space
                for the taget ship extraction process.
                AIS records outside this BoundingBox are 
                not eligible for TargetShip construction.
        dynamic_paths: path or list of paths to a csv file 
                       containing AIS messages of type 1,2,3 and 18.
        static_paths: path or list of paths to a csv file 
                      containing AIS messages of type 5.
        max_tgt_ships: maximum number of target ships to retrun

        preproceccor: Preprocessor function for input data. 
                Defaults to the identity.
        decoded: boolean flag indicating whether the input data
                is already decoded or not. If `decoded` is False,
                the input data is decoded on the fly. Please note
                that this process is magnitudes slower than using
                decoded data.
        high_accuracy: boolean flag indicating whether Vincenty's
                formula should be used for distance calculations.
                If `high_accuracy` is False, the Haversine formula
                is used. Defaults to False.
        """

        if not isinstance(dynamic_paths,list):
            dynamic_paths = [dynamic_paths]
            dynamic_paths = [Path(p) for p in dynamic_paths]
        else:
            dynamic_paths = [Path(p) for p in dynamic_paths]

        if not isinstance(static_paths,list):
            static_paths = [static_paths]
            static_paths = [Path(p) for p in static_paths]
        else:
            static_paths = static_paths
            static_paths = [Path(p) for p in static_paths]
            
        self.data_loader = DataLoader(
            dynamic_paths,
            static_paths,
            preprocessor
        )

        # Spatial bounding box of current AIS message space
        self.FRAME = frame
        
        self.high_accuracy = high_accuracy
        
        self.spatial_filter = (
            f"{Msg12318Columns.LON.value} > {frame.LONMIN} and "
            f"{Msg12318Columns.LON.value} < {frame.LONMAX} and "
            f"{Msg12318Columns.LAT.value} > {frame.LATMIN} and "
            f"{Msg12318Columns.LAT.value} < {frame.LATMAX}"
        )

        # Maximum number of target ships to extract
        self.max_tgt_ships = max_tgt_ships
        
        # Maximum temporal deviation of target
        # ships from provided time in `init()`.
        # This is used internally to catch
        # all messages around the queried time,
        # for building and interpolating trajectories.
        self.time_delta = 30 # in minutes

        self.neighborhood = NeighborhoodTreeSearch(
            self.data_loader,
            self.time_delta,
            self.max_tgt_ships
        )
        
        self.constructor = TargetShipConstructor(
            self.data_loader
        )
        
        # Flag indicating whether the input data
        # is already decoded or not
        self.decoded = decoded
        
        # Length of the original AIS message dataframe
        self._n_original = 0
        # Length of the filtered AIS message dataframe
        self._n_filtered = 0
        
        # Number of times, the _speed_correction function
        # was called
        self._n_speed_correction = 0
        
        # Number of times, the _position_correction function
        # was called
        self._n_position_correction = 0
        
    def freeze(self, 
               tpos: TimePosition, 
               search_radius: float = 20, # in nautical miles
               interpolation: str = "linear") -> Targets:
        """
        Freeze around a given position and time and return 
        a list of target ships present in the neighborhood 
        around the given position. 
        
        tpos: TimePosition object of agent for which 
                neighbors shall be found
        
        search_radius: Radius around agent in which taget vessel 
                search is rolled out [nm]

        overlap_tpos: boolean flag indicating whether
                we only want to return vessels, whose track overlaps
                with the queried timestamp. If `overlap_tpos` is False,
                all vessels whose track is within the time delta of the
                queried timestamp are returned.
        """
        assert interpolation in ["linear","spline","auto"], \
            "Interpolation method must be either 'linear', 'spline' or 'auto'"
        # Get neighbors
        neighbors = self.neighborhood.get_neighbors(tpos,search_radius)
        tgts = self.constructor._sp_construct_target_vessels(neighbors,tpos,True)
        # Contruct Splines for all target ships
        tgts = self._interpolate_trajectories(tgts,mode=interpolation)
        return tgts
    
    def extract_all(self,
                    njobs: int = 4, 
                    skip_tsplit: bool = False) -> Targets:
        """
        Extracts a dictionary of all target ships in the
        frame of the search agent using the split-point
        method as described in the paper. 
        
        Parameters:
        ----------
        njobs: int
            Number of jobs to use for multiprocessing.
        skip_tsplit: bool
            If True, the split-point method is not used
            to split the tracks of the target ships.
            Instead, all vessels will only have one track,
            containing all time-ordered AIS messages
            comning from the same MMSI.
            
        Returns:
        --------
        dict: A dictionary of dict[MMSI,TargetShip]. 
        """
        return self.constructor._mp_construct_target_vessels(
            njobs,skip_tsplit
        )

    def _interpolate_trajectories(self, 
                           tgts: Targets,
                           mode: str = "auto") -> Targets:
        """
        Interpolate all target ships' trajectories.
        """
        for mmsi,tgt in list(tgts.items()):
            try:
                tgt.interpolate(mode) # Construct splines
            except InterpolationError as e:
                logger.warn(e)
                del tgts[mmsi]
        return tgts


class TargetShipConstructor:
    """
    A class for constructing TargetShip 
    objects from raw AIS messages.

    Attributes:
        `data_loader`: 
            An instance of the `pytsa.utils.DataLoader` 
            class used to load AIS message data.

    Methods:
        `construct_target_vessels`: 
            Constructs TargetVessel objects from 
            a given DataFrame.
    """
    def __init__(self, data_loader: DataLoader) -> None:
        
        self.data_loader = data_loader


    def _distribute(self, df: pd.DataFrame) -> list[pd.DataFrame]:
        """
        Distribute a given dataframe into smaller
        dataframes, each containing only messages
        from a single MMSI.
        """
        mmsis = df[Msg12318Columns.MMSI.value].unique()
        return [df[df[Msg12318Columns.MMSI.value] == mmsi] for mmsi in mmsis]
    
    def _impl_construct_target_vessel(self, 
                                      dyn: pd.DataFrame,
                                      stat: pd.DataFrame) -> Targets:
        """
        Construct a single TargetVessel object from a given dataframe.
        """
        MMSI = int(dyn[Msg12318Columns.MMSI.value].iloc[0])
        # Initialize TargetVessel object
        tv =  TargetShip(
            ts = None,
            mmsi=MMSI,
            ship_type=self._get_ship_type(stat,MMSI),
            length=self._get_ship_length(stat,MMSI),
            tracks=[[]]
        )
        first = True
                
        dyn = dyn.sort_values(by=BaseColumns.TIMESTAMP.value)
        
        for ts,lat,lon,sog,cog in zip(
            dyn[BaseColumns.TIMESTAMP.value], dyn[Msg12318Columns.LAT.value],  
            dyn[Msg12318Columns.LON.value],       dyn[Msg12318Columns.SPEED.value],
            dyn[Msg12318Columns.COURSE.value]):
            ts: pd.Timestamp # make type hinting happy
            
            msg = AISMessage(
                sender=MMSI,
                timestamp=int(ts.timestamp()), # Convert to unix
                lat=lat,lon=lon,
                COG=cog,SOG=sog
            )
            if first:
                tv.tracks[-1].append(msg)
                first = False
            else:
                # Check if message is a duplicate, i.e. had been 
                # received by multiple AIS stations
                if msg.timestamp == tv.tracks[-1][-1].timestamp:
                    continue
                tv.tracks[-1].append(msg)

        return {MMSI:tv}
    
    def _mp_construct_target_vessels(self,
                                     njobs: int = 4,
                                     skip_tsplit: bool = False) -> Targets:
        """
        Adaption from `_construct_target_vessels` with multiprocessing.
        The initial dataframe is split into smaller dataframes, each
        containing only messages from a single MMSI. These dataframes
        are then processed in parallel and the results are merged.
        """
        singles: list[Targets] = []
        with mp.Pool(njobs) as pool:
            for dyn, stat in self.data_loader.iterate_chunks():
                single_frames = self._distribute(dyn)
                res = pool.starmap(
                    self._impl_construct_target_vessel,
                    [(dframe,stat) for dframe in single_frames]
                )
                singles.extend(res)
        targets = self._join_tracks(*singles)
        targets = self._rm_dups(targets)
        if not skip_tsplit:
            targets = self._determine_split_points(targets)
        return targets
        
    def _rm_dups(self, targets: Targets) -> Targets:
        """
        Remove duplicate positions and timestamps.
        Duplicate positions are removed by the 
        set operation, as equality of AISMessage
        objects is defined by their lat/lon.
        
        Duplicate timestamps are removed by sorting
        the list of AISMessage objects by their
        timestamp and then removing all but the
        first occurence of a timestamp.
        """
        for tgt in targets.values():
            # Rm pos dups
            rmpos = list(set(tgt.tracks[0]))
            # Check for timestamp dups
            ts_sorted_idx = np.unique(
                [msg.timestamp for msg in rmpos],
                return_index=True)[1]
            tgt.tracks[0] = [rmpos[i] for i in ts_sorted_idx]
            
        return targets
    
    def _join_tracks(self,
                    *singles: Targets) -> Targets:
        """
        Join tracks of multiple TargetShip objects
        into a single TargetShip object.
        
        Note: The input TargetShip objects are only 
        consisting of a single track each, as they
        just have been constructed.
        """
        targets: Targets = {}
        for single in singles:
            for mmsi,tgt in single.items():
                if mmsi not in targets:
                    targets[mmsi] = tgt
                else:
                    targets[mmsi].tracks[0].extend(
                        tgt.tracks[0]
                    )
        return targets
                    
    def _determine_split_points(self,
                                targets: Targets) -> Targets:
        """
        Determine split points for all target ships.
        """
        nvessels = len(targets)
        ctr = 0
        for tgt in list(targets.values()):
            logger.info(
                f"Processing target ship {tgt.mmsi} "
                f"({ctr+1}/{nvessels})"
            )
            ctr += 1
            for track in tgt.tracks:
                track.sort(key=lambda x: x.timestamp)
                _itracks = [] # Intermediary track
                tstartidx = 0
                for i, (msg_t0,msg_t1) in enumerate(pairwise(track)):
                    if split.is_split_point(msg_t0,msg_t1):
                        _itracks.append(track[tstartidx:i+1])
                        tstartidx = i+1
            # Only keep tracks with more than one observation
            tgt.tracks = [track for track in _itracks if len(track) > 1]
            # If no tracks are left, remove target ship
            if not tgt.tracks:
                logger.warning(
                    f"Target ship {tgt.mmsi} has no tracks left after filtering."
                )
                del targets[tgt.mmsi]
        logger.info("Splitting done.")
        return targets
                        
    def _sp_construct_target_vessels(self,
                                     neighbors: pd.DataFrame,
                                     tpos: TimePosition,
                                     overlap: bool) -> Targets:
        """
        Single-process version of `_construct_target_vessels`.
        
        Walk through the rows of `df` and construct a 
        `TargetVessel` object for every unique MMSI. 
        
        The individual AIS Messages are sorted by date
        and are added to the respective TargetVessel's track attribute.
        
        The `overlap` flag indicates whether we only want to return vessels,
        whose track overlaps with the queried timestamp. If `overlap` is False,
        all vessels whose track is within the time delta of the
        queried timestamp are returned.
        
        If the temporal or spatial difference between two AIS Messages is too large,
        the track of the target ship is split into two tracks.
        
        """
        targets: Targets = {}
        neighbors = neighbors.sort_values(by=BaseColumns.TIMESTAMP.value)
        
        for mmsi,ts,lat,lon,sog,cog in zip(
            neighbors[Msg12318Columns.MMSI.value], neighbors[BaseColumns.TIMESTAMP.value],
            neighbors[Msg12318Columns.LAT.value],  neighbors[Msg12318Columns.LON.value],
            neighbors[Msg12318Columns.SPEED.value],neighbors[Msg12318Columns.COURSE.value]):
            ts: pd.Timestamp # make type hinting happy
            
            msg = AISMessage(
                sender=mmsi,
                timestamp=ts.to_pydatetime().timestamp(), # Convert to unix
                lat=lat,lon=lon,
                COG=cog,SOG=sog
            )
            
            if mmsi not in targets:
                targets[mmsi] = TargetShip(
                    ts = tpos.timestamp if tpos is not None else None,
                    mmsi=mmsi,
                    ship_type=self._get_ship_type(self.data_loader.static_data,mmsi),
                    length=self._get_ship_length(self.data_loader.static_data,mmsi),
                    tracks=[[msg]]
                )
            else:
                # Split track if change in speed or heading is too large
                if split.is_split_point(targets[mmsi].tracks[-1][-1],msg):
                    targets[mmsi].tracks.append([])
                v = targets[mmsi]
                v.tracks[-1].append(msg)

        for tgt in targets.values():
            tgt.find_shell() # Find shell (start/end of traj) of target ship
        
        # Remove target ships with only one observation
        self._remove_single_obs(targets)
        
        if overlap:
            self._filter_overlapping(targets,tpos)
            
        return targets
    
    def _filter_overlapping(self, 
                            targets: Targets, 
                            tpos: TimePosition) -> None:
        """
        Remove target ships whose track does not overlap
        with the queried timestamp.
        """
        for tgt in targets.values():
            to_keep = []
            for track in tgt.tracks:
                if self._overlaps_search_date(track,tpos):
                    to_keep.append(track)
            tgt.tracks = to_keep

    def _remove_single_obs(self, targets: Targets) -> Targets:
        """
        Remove tracks that only have a single observation.
        """
        for tgt in targets.values():
            to_keep = []
            for track in tgt.tracks:
                if len(track) >= 2:
                    to_keep.append(track)
            tgt.tracks = to_keep
    
    def _break_down_velocity(self,
                             speed: float,
                             course: float) -> tuple[float,float]:
        """
        Break down a given velocity into its
        longitudinal and lateral components.
        """
        return (
            speed * np.cos(np.deg2rad(course)), # Longitudinal component
            speed * np.sin(np.deg2rad(course)) # Lateral component
        )
    
    def _overlaps_search_date(self, 
                              track: Track, 
                              tpos: TimePosition) -> bool:
        """
        Return True if the track of the given vessel
        overlaps with the queried timestamp.
        """
        return (track[0].timestamp < tpos.timestamp < track[-1].timestamp)

    def _get_ship_type(self, 
                       static_msgs: pd.DataFrame,
                       mmsi: int) -> list[int]:
        """
        Return the ship type of a given MMSI number.

        If more than one ship type is found, the first
        one is returned and a warning is logged.
        """
        st = static_msgs[static_msgs[Msg5Columns.MMSI.value] == mmsi]\
            [Msg5Columns.SHIPTYPE.value].values
        st:np.ndarray = np.unique(st)
        if st.size > 1:
            logger.warning(
                f"More than one ship type found for MMSI {mmsi}. "
                f"Found {st}.")
            return st
        else:
            return list(st)

    def _get_ship_length(self, 
                         static_msgs: pd.DataFrame,
                         mmsi: int) -> int:
        """
        Return the ship length of a given MMSI number.

        If more than one ship length is found, the largest
        one is returned and a warning is logged.
        """
        raw = static_msgs[static_msgs[Msg5Columns.MMSI.value] == mmsi]\
            [[Msg5Columns.TO_BOW.value,Msg5Columns.TO_STERN.value]].values
        sl:np.ndarray = np.sum(raw,axis=1)
        sl = np.unique(sl)
        if sl.size > 1:
            logger.warning(
                f"More than one ship length found for MMSI {mmsi}. "
                f"Found {sl}. Returning {max(sl)}.")
            return max(sl)
        return sl



"""
Module for visualizing both the ECDFs of the split point metrics,
and the extracted trajectories.

This init file handles the creation of the `plots`
folder, where the plots are saved, houses
helper functions for the plotting functions
and defines the default plot aesthetics.
"""
from typing import Optional
from pathlib import Path
import matplotlib.pyplot as plt
import matplotlib as mpl
import colorsys

# Plotting folder -------------------------------------------
def _create_default_plot_folder() -> Path:
    """
    Creates the default plot folder.
    """
    path = Path("plots")
    if not path.exists():
        path.mkdir()
    return path

def register_plot_dir(path: Optional[str]) -> None:
    """
    Change the default plot folder.
    """
    global PLOT_FOLDER
    path = Path(path)
    if not path.exists():
        path.mkdir()
    PLOT_FOLDER = path

# Export the default plot folder
PLOT_FOLDER = _create_default_plot_folder()

# Plotting helpers ------------------------------------------
cc = mpl.colors.ColorConverter.to_rgb

def scale_lightness(rgb, scale_l):
    """
    Scale the lightness of an rgb color.
    """
    # Convert rgb to hls (hue, lightness, saturation)
    h, l, s = colorsys.rgb_to_hls(*rgb)
    # Manipulate h, l, s values and return as rgb
    return colorsys.hls_to_rgb(h, min(1, l * scale_l), s = s)

# Colorwheels-----------------------------------------------
COLORWHEEL = [
    "#264653",
    "#2a9d8f",
    "#e9c46a",
    "#f4a261",
    "#e76f51",
    "#E45C3A",
    "#732626"
]
COLORWHEEL_DARK = [scale_lightness(cc(c), 0.6) for c in COLORWHEEL]

COLORWHEEL2 = [
    "#386641", 
    "#6a994e", 
    "#a7c957", 
    "#f2e8cf", 
    "#bc4749"
]
COLORWHEEL2_DARK = [scale_lightness(cc(c), 0.6) for c in COLORWHEEL2]

COLORWHEEL3 = [
    "#335c67",
    "#fff3b0",
    "#e09f3e",
    "#9e2a2b",
    "#540b0e"
]
COLORWHEEL_MAP = [
    "#0466c8",
    "#0353a4",
    "#023e7d",
    "#002855",
    "#001845",
    "#001233",
    "#33415c",
    "#5c677d",
    "#7d8597",
    "#979dac"
]

# Matplotlib settings ---------------------------------------
plt.style.use('bmh')
plt.rcParams["font.family"] = "monospace"


"""
This module aims at finding all spatio-temporal
neighbors around a given (stripped) AIS Message. 

A provided geographical area will be split into 
evenly-sized grids
"""
from __future__ import annotations

from datetime import datetime
from typing import List, Union

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline, interp1d
from ..structs import AISMessage, MMSI, Track
from ..logger import logger

# Type aliases
Latitude = float
Longitude = float

# Constants
PI = np.pi


# Exceptions
class OutofTimeBoundsError(Exception):
    pass
class InterpolationError(Exception):
    pass

class TrackSplines:
    """
    Class for performing separate univariate
    spline interpolation on a given AIS track.

    This class provides the following attributes:
        - lat: Spline interpolation of latitude
        - lon: Spline interpolation of longitude
        - northing: Spline interpolation of northing
        - easting: Spline interpolation of easting
        - COG: Spline interpolation of course over ground
        - SOG: Spline interpolation of speed over ground
        - ROT: Spline interpolation of rate of turn
        - dROT: Spline interpolation of change of rate of turn

        All splines are univariate splines, with time as the
        independent variable.
    """
    def __init__(self, track: List[AISMessage]) -> None:
        self.track = track
        self._attach_splines()

    def _attach_splines(self) -> None:
        """
        Perform spline interpolation on the
        AIS track and attach the results to
        the class as attributes.
        """
        timestamps = [int(msg.timestamp) for msg in self.track]

        self.lat = InterpolatedUnivariateSpline(
            timestamps, [msg.lat for msg in self.track]
        )
        self.lon = InterpolatedUnivariateSpline(
            timestamps, [msg.lon for msg in self.track]
        )
        # Add the 360° complement to correctly interpolate
        # course changes from 360° to 0°
        complement360 = np.rad2deg(
            np.unwrap(
                np.deg2rad([msg.COG for msg in self.track])
            )
        )
        self.COG = InterpolatedUnivariateSpline(
            timestamps, complement360
        )
        self.SOG = InterpolatedUnivariateSpline(
            timestamps, [msg.SOG for msg in self.track]
        )

class TrackLinear:
    """
    Linear interpolation of a given AIS track.
    """
    def __init__(self, track: List[AISMessage]) -> None:
        self.track = track
        self._attach_linear()

    def _attach_linear(self) -> None:
        """
        Perform linear interpolation on the
        AIS track and attach the results to
        the class as attributes.
        """
        timestamps = [msg.timestamp for msg in self.track]
        
        self.lat = interp1d(
            timestamps, [msg.lat for msg in self.track]
        )
        self.lon = interp1d(
            timestamps, [msg.lon for msg in self.track],
            
        )
        # Add the 360° complement to correctly interpolate
        # course changes from 360° to 0°
        complement360 = np.rad2deg(
            np.unwrap(
                np.deg2rad([msg.COG for msg in self.track])
            )
        )
        self.COG = interp1d(
            timestamps, complement360,
            bounds_error=False, fill_value=None
        )
        self.SOG = interp1d(
            timestamps, [msg.SOG for msg in self.track]
        )
class TargetShip:
    """
    Central object for the pytsa package. It holds
    a collection of functions for working with the 
    AIS messages belonging to it.
    
    Attributes:
        - ts: Timestamp for which to observe the vessel
        - mmsi: Maritime Mobile Service Identity
        - track: List of AIS messages belonging to the
                target vessel
        - ship_type: Ship type of the target vessel
        - length: Length of the target vessel
                
    Methods:
        - interpolate(): Construct smoothed univariate splines
                for northings, eastings, course over ground,
                speed over ground, rate of turn and change of
                rate of turn contained in the vessels' track.
                
                If the caller wants to use linear interpolation
                instead of splines, the lininterp flag can be set
                to True.
                
        - observe_at_query(): Returns the 6-tuple of [northing,
                easting, COG, SOG, ROT, dROT] at the timestamp 
                the object was initialized with.
                The timestamp should lie within the
                track's temporal bounds.
                
                Currently, there is no check 
                for the timestamp being within the track's
                temporal bounds, which can lead to errors.
                If you want to use this function, please
                make sure that the timestamp is within bounds.
                
                You can also pass a timestamp to the function
                to observe the vessel at a different timestamp.
                Please note that there is still no check for
                the timestamp being within the track's bounds.
                
        - observe_interval(): Returns an array of 7-tuples of [northing,
                easting, COG, SOG, ROT, dROT, timestamp] for 
                the track between the given start and end timestamps,
                with the given interval in [seconds].
                
        - fill_rot(): Fill out missing rotation data and first
                derivative of rotation by inferring it from the
                previous and next AIS messages' headings.
        
        - overwrite_rot(): Overwrite the ROT and dROT values with
                the values from COG.
        
        - find_shell(): Find the two AIS messages encompassing
                the objects' track elements and save them
                as attributes.
        
        - ts_to_unix(): Convert the vessel's timestamp for
                each track element to unix time.
    """    
    
    def __init__(
        self,
        ts: Union[int, None], 
        mmsi: MMSI, 
        tracks: List[List[AISMessage]],
        ship_type: list[int] = None,
        length: float = None,
        ) -> None:
        
        self.ts = ts
        self.mmsi = mmsi 
        self.tracks = tracks
        self.ship_type = ship_type
        self.length = length
        self.lininterp = False # Linear interpolation flag
        
        # Indicate if the interpolation function have 
        # been attached to the object
        self.interpolation: list[TrackSplines] | list[TrackLinear] | None = None
    
    def interpolate(self,mode: str) -> None:
        """
        Construct splines for the target vessel
        """
        assert mode in ["linear","spline","auto"],\
            "Mode must be either 'linear', 'spline' or 'auto'."
        if not self.tracks:
            raise InterpolationError(
                f"Empty track for vessel {self.mmsi}."
            )
        self.interpolation = []
        for track in self.tracks:
            try:
                if mode == "auto" and self.lininterp:
                    self.interpolation.append(TrackLinear(track))
                elif mode == "auto" and not self.lininterp:
                        self.interpolation.append(TrackSplines(track))
                elif mode == "linear":
                    self.interpolation.append(TrackLinear(track))
                elif mode == "spline":
                    self.interpolation.append(TrackSplines(track))
            except Exception as e:
                logger.error(
                    f"Could not interpolate track "
                    f"of vessel {self.mmsi}: {e}."
                )
                self.tracks.remove(track)
        if not self.interpolation: 
            raise InterpolationError(
                f"Could not interpolate any trajecotry "
                f"of vessel {self.mmsi}."
            )

    def observe(self) -> np.ndarray:
        """
        Infers
            - Latitude,
            - Longitude,
            - Course over ground (COG) [degrees],
            - Speed over ground (SOG) [knots],
            
        from an interpolated trajectory for the timestamp, 
        the object was initialized with.

        """
        if self.ts is None:
            msg = (
                "`observe_at()` method not available for use with `get_all_ships()`.\n"
                "Please use `observe_interval()` or instatiante the object through `get_ships()`." 
                )
            logger.error(msg)
            raise NotImplementedError(msg)
        assert self.interpolation is not None,\
            "Interpolation has not been run. Call interpolate() first."
        # Convert query timestamp to unix time
        if isinstance(self.ts, datetime):
            ts = self.ts.timestamp()
        else: ts = self.ts
            
        # Check if the query timestamp is within the
        # track's timestamps
        for i,track in enumerate(self.tracks):
            if self._is_in_interval(ts,track):
                break
            if i == len(self.tracks)-1:
                raise OutofTimeBoundsError(
                    "Query timestamp is not within the track's timestamps."
                )

        # Return the observed values from the splines
        # at the given timestamp
        # Returns a 1x4 array:
        # [northing, easting, COG, SOG]
        return np.array([
            self.interpolation[i].lat(ts),
            self.interpolation[i].lon(ts),
            # Take the modulo 360 to ensure that the
            # course over ground is in the interval [0,360]
            self.interpolation[i].COG(ts) % 360,
            self.interpolation[i].SOG(ts)
        ])

    def observe_interval(
            self, 
            start: datetime, 
            end: datetime, 
            interval: int
            ) -> np.ndarray:
        """
        Infers
            - Northing (meters),
            - Easting (meters),
            - Course over ground (COG) [degrees],
            - Speed over ground (SOG) [knots],

        from univariate splines for the track between
        the given start and end timestamps, with the
        given interval in [seconds].
        """
        assert self.interpolation is not None,\
            "Interpolation has not been run. Call interpolate() first."
        # Convert query timestamps to unix time
        if isinstance(start, datetime):
            start = start.timestamp()
            end = end.timestamp()

        # Check if the start timestamp is within the
        # track's timestamps
        for i,track in enumerate(self.tracks):
            if self._is_in_interval(start,track):
                break
            if i == len(self.tracks)-1:
                raise OutofTimeBoundsError(
                    "Start timestamp is not within the track's timestamps. "
                    "MMSI: {}".format(self.mmsi)
                )
        # Check if the end timestamp is within the
        # track's timestamps
        for j,track in enumerate(self.tracks):
            if self._is_in_interval(end,track):
                break
            if j == len(self.tracks)-1:
                raise OutofTimeBoundsError(
                    "End timestamp is not within the track's timestamps. "
                    "MMSI: {}".format(self.mmsi)
                )

        # Create a list of timestamps between the start and end
        # timestamps, with the given interval
        timestamps = np.arange(start, end, interval)

        # Start and end points of the interval
        # lie within the same track
        if i == j:
            # Return the observed values from the interpolation
            # at the given timestamps
            # Returns a Nx4 array:
            # [northing, easting, COG, SOG, timestamp]
            preds: np.ndarray = np.array([
                self.interpolation[i].lat(timestamps),
                self.interpolation[i].lon(timestamps),
                # Take the modulo 360 of the COG to get the
                # heading to be between 0 and 360 degrees
                self.interpolation[i].COG(timestamps) % 360,
                self.interpolation[i].SOG(timestamps),
                timestamps
            ])
            return preds.T
        # Start and end points of the interval
        # lie in different tracks
        else:
            logger.warning(
                "Start and end points of the interval lie in different tracks. "
                "Results may be inaccurate. MMSI: %s".format(self.mmsi)
            )
            # Combine the tracks into one list
            tracks = self.tracks[i:j+1]
            tracks = [msg for track in tracks for msg in track]
            interp_ = TrackSplines(tracks) if not self.lininterp else TrackLinear(tracks)
            # Return the observed values from the interpolation
            # at the given timestamps
            # Returns a Nx4 array:
            # [northing, easting, COG, SOG, timestamp]
            preds: np.ndarray = np.array([
                interp_.lat(timestamps),
                interp_.lon(timestamps),
                # Take the modulo 360 of the COG to get the
                # heading to be between 0 and 360 degrees
                interp_.COG(timestamps) % 360,
                interp_.SOG(timestamps),
                timestamps
            ])
            return preds.T
            
    
    def _is_in_interval(self, query: int, msgs: Track) -> bool:
        """
        Check if the query timestamp is within the
        given interval.
        """
        return (query >= msgs[0].timestamp) and (query <= msgs[-1].timestamp)


    def find_shell(self) -> None:
        """
        Find the two AIS messages encompassing
        the objects' track elements and save them
        as attributes.
        """ 
        self.lower, self.upper = [], []
        for track in self.tracks:
            self.lower.append(track[0])
            self.upper.append(track[-1])

Targets = dict[MMSI,TargetShip]
class TrajectoryMatcher:
    """
    Class for matching trajectories of two vessels.

    The trajectories are matched if they overlap
    for a given threshold.

    The trajectories are disjoint if they do not
    overlap at all.
    """

    def __init__(
            self, 
            vessel1: TargetShip, 
            vessel2: TargetShip,
            threshold: float = 20 # [min] Minimun overlap between trajectories
            ) -> None:
        self.vessel1 = vessel1
        self.vessel2 = vessel2

        self.vessel1.find_shell()
        self.vessel2.find_shell()

        if self._disjoint():
            self.disjoint_trajectories = True
            self.overlapping_trajectories = False
        else:
            self.disjoint_trajectories = False
            self._start()
            self._end()
            if self._overlapping(threshold):
                self.overlapping_trajectories = True
            else:
                self.overlapping_trajectories = False

    def _overlapping(self, threshold: int) -> bool:
        """
        Check if the trajectories overlap for a given threshold
        """
        return (self.end - self.start) > threshold * 60
    
    def _start(self) -> None:
        """
        Find the starting point included in both trajectories
        """
        if self.vessel1.lower.timestamp < self.vessel2.lower.timestamp:
            self.start = self.vessel2.lower.timestamp
        else:
            self.start = self.vessel1.lower.timestamp

    def _end(self) -> None:
        """
        Find the end point included in both trajectories
        """
        if self.vessel1.upper.timestamp > self.vessel2.upper.timestamp:
            self.end = self.vessel2.upper.timestamp
        else:
            self.end = self.vessel1.upper.timestamp
    
    def _disjoint(self) -> bool:
        """
        Check if the trajectories are disjoint on the time axis
        """
        return (
            (self.vessel1.upper.timestamp < self.vessel2.lower.timestamp) or
            (self.vessel2.upper.timestamp < self.vessel1.lower.timestamp) or 
            (self.vessel1.lower.timestamp > self.vessel2.upper.timestamp) or
            (self.vessel2.lower.timestamp > self.vessel1.upper.timestamp)
        )

    def observe_interval(self,interval: int) -> TrajectoryMatcher:
        """
        Retruns the trajectories of both vessels
        between the start and end points, with the
        given interval in [seconds].
        """
        if self.disjoint_trajectories:
            raise ValueError(
                "Trajectories are disjoint on the time scale."
            )
        
        obs_vessel1 = self.vessel1.observe_interval(
            self.start, self.end, interval
        )
        obs_vessel2 = self.vessel2.observe_interval(
            self.start, self.end, interval
        )

        self.obs_vessel1 = obs_vessel1
        self.obs_vessel2 = obs_vessel2
        
        return self

def _dtr(angle: float) -> float:
    """Transform from [0,360] to 
    [-180,180] in radians"""
    o = ((angle-180)%360)-180
    return o/180*np.pi

def _dtr2(angle: float) -> float:
    """Transform from [0,360] to 
    [-180,180] in radians and switch
    angle rotation order since python
    rotates counter clockwise while
    heading is provided clockwise"""
    o = ((angle-180)%360)-180
    return -o/180*np.pi



"""
Utility functions for pytsa
"""
from datetime import datetime
from itertools import cycle
import math
from pathlib import Path
from threading import Thread
import time
from typing import Callable, Generator
import pandas as pd
import psutil
import vincenty as _vincenty
import ciso8601
from .logger import logger
from .decode import decode_from_file
from .decode.filedescriptor import (
    BaseColumns, Msg12318Columns, Msg5Columns   
)

def m2nm(m: float) -> float:
    """Convert meters to nautical miles"""
    return m/1852

def nm2m(nm: float) -> float:
    """Convert nautical miles to meters"""
    return nm*1852

def s2h(s: float) -> float:
    """Convert seconds to hours"""
    return s/3600

def mi2nm(mi: float) -> float:
    """Convert miles to nautical miles"""
    return mi/1.151

def vincenty(lon1, lat1, lon2, lat2, miles = True) -> float:
    """
    Calculate the great circle distance in kilometers between two points 
    on the earth (specified in decimal degrees)
    """
    p1 = (lat1,lon1)
    p2 = (lat2,lon2)
    return _vincenty.vincenty_inverse(p1,p2,miles=miles)

def haversine(lon1, lat1, lon2, lat2, miles = True):
    """
    Calculate the great circle distance in kilometers between two points 
    on the earth (specified in decimal degrees)
    """
    # convert decimal degrees to radians 
    lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])

    # haversine formula 
    dlon = lon2 - lon1 
    dlat = lat2 - lat1 
    a = (
        math.sin(dlat/2)**2 + math.cos(lat1) * 
        math.cos(lat2) * math.sin(dlon/2)**2
        )
    c = 2 * math.asin(math.sqrt(a)) 
    r = 3956 if miles else 6371 # Radius of earth in kilometers or miles
    return c * r

def greater_circle_distance(lon1, lat1, lon2, lat2, miles = True, method = "haversine"):
    """
    Calculate the great circle distance in kilometers between two points 
    on the earth (specified in decimal degrees)
    """
    assert method in ["haversine", "vincenty"], "Invalid method: {}".format(method)
    if method == "haversine":
        return haversine(lon1, lat1, lon2, lat2, miles=miles)
    return vincenty(lon1, lat1, lon2, lat2, miles=miles)

def heading_change(h1,h2):
    """
    calculate the change between two headings
    such that the smallest angle is returned.
    """

    diff = abs(h1-h2)
    if diff > 180:
        diff = 360 - diff
    if (h1 + diff) % 360 == h2:
        return diff
    else:
        return -diff
    
class DataLoader:
    """
    Data loader for AIS data.
    
    Takes a list of paths to static and dynamic data files
    provides utilities to decode and load the data.
    """
    
    # Fraction of the file to load
    chunkfrac = 1/3
    
    dynamic_columns = [
        BaseColumns.TIMESTAMP.value,
        Msg12318Columns.MMSI.value,
        Msg12318Columns.LAT.value,
        Msg12318Columns.LON.value,
        Msg12318Columns.SPEED.value,
        Msg12318Columns.COURSE.value
    ]
    
    static_columns = [
        Msg5Columns.MMSI.value,
        Msg5Columns.SHIPTYPE.value,
        Msg5Columns.TO_BOW.value,
        Msg5Columns.TO_STERN.value
    ]
    
    def __init__(self, 
                 dynamic_paths: list[Path],
                 static_paths: list[Path],
                 pre_processor: Callable[[pd.DataFrame],pd.DataFrame]) -> None:
        
        self.preprocessor = pre_processor
        
        # Check that we only have csv files
        assert all(
            [f.suffix == ".csv" for f in dynamic_paths + static_paths]
        ), "Only csv files are supported."
        
        # Check for same size and same order
        self.sdyn, self.sstat = self.align_data_files(
            dynamic_paths, static_paths
        )
        
        # Flag to indicate if the data has
        # been loaded into memory entirely.
        # Only used for the `load` method.
        self.loaded = False
    
    @staticmethod
    def _date_transformer(datefile: Path) -> datetime:
        """
        Transform a date string in the format YYYY_MM_DD
        to a datetime object.
        """
        return ciso8601.parse_datetime(datefile.stem.replace("_", "-"))

    @staticmethod
    def align_data_files(dyn: list[Path], 
                         stat: list[Path]
                         ) -> tuple[list[Path], list[Path]]:
        """
        Align dynamic and static data files by sorting them
        according to their date, and removing all files that
        are not present in both lists.

        This function assumes that the dynamic and static data files
        are named according to the following convention:
        - dynamic data files: YYYY_MM_DD.csv
        - static data files: YYYY_MM_DD.csv
        
        In case your input data files are not named according to this
        convention, you are advised to either rename them accordingly
        or adapt the `_date_transformer` function, which is used to
        sort the files by date.
        """
        if len(dyn) != len(stat):

            print(
                "Number of dynamic and static messages do not match."
                f"Dynamic: {len(dyn)}, static: {len(stat)}\n"
                "Processing only common files."
            )
            # Find the difference
            d = set([f.stem for f in dyn])
            s = set([f.stem for f in stat])
            
            # Find all files that are in d and s
            common = list(d.intersection(s))
            
            # Remove all files that are not in common
            dyn = [f for f in dyn if f.stem in common]
            stat = [f for f in stat if f.stem in common]
            
        # Sort the files by date
        dyn = sorted(dyn, key=DataLoader._date_transformer)
        stat = sorted(stat, key=DataLoader._date_transformer)

        assert all([d.stem == s.stem for d,s in zip(dyn, stat)]),\
            "Dynamic and static messages are not in the same order."

        return dyn, stat
    
    def _dynamic_preprocessor(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Preprocess the dynamic messages.
        """
        # Apply custom filter
        df = self.preprocessor(df).copy()
        # Convert timestamp to datetime
        df[BaseColumns.TIMESTAMP.value] = pd.to_datetime(
                df[BaseColumns.TIMESTAMP.value])
        # Drop duplicates form multiple base stations
        df = df.drop_duplicates(
                subset=[
                    BaseColumns.TIMESTAMP.value,
                    Msg12318Columns.MMSI.value
                ], keep="first"
            )
        return df
    
    def from_raw(self, raw_dyn: Path, raw_stat: Path) -> tuple[pd.DataFrame,pd.DataFrame]:
        """
        Return a DataFrame containing the decoded dynamic and static
        messages.
        
        Decoding the raw AIS messages will result in a 
        DataFrame containing all fields of a NMEA message.
        Since we are only interested in a subset of the fields,
        we select only those fields and return them as a new
        DataFrame.
        
        Note that the file is NOT decoded in chunks, but rather
        as a whole.
        """
        # Decode dynamic data
        dyn = decode_from_file(raw_dyn, None, save_to_file=False)
        dyn = self._dynamic_preprocessor(dyn)
        # Decode static data
        stat = decode_from_file(raw_stat, None, save_to_file=False)
        return (
            dyn[[DataLoader.dynamic_columns]], 
            stat[[DataLoader.static_columns]]
        )
        
    def load(self) -> None:
        """
        Loads all data into memory.
        """
        self.dynamic_data = pd.DataFrame()
        self.static_data = pd.DataFrame()
        for dyn_path, stat_path in zip(self.sdyn,self.sstat):
            logger.info(f"Loading {stat_path.stem} and {dyn_path.stem}")
            d = pd.read_csv(dyn_path,sep=",",usecols=self.dynamic_columns)
            d = self._dynamic_preprocessor(d)
            s = pd.read_csv(stat_path,sep=",",usecols=self.static_columns)
            self.dynamic_data = pd.concat([self.dynamic_data,d])
            self.static_data = pd.concat([self.static_data,s])
        logger.info("Done.")
        
        self.loaded = True
    
    def iterate_chunks(self,
                   decode: bool = False
                ) -> Generator[
                      tuple[pd.DataFrame,pd.DataFrame], 
                      None, 
                      None
                    ]:
        """
        Yield chunks of the dynamic and static
        messages.
        
        If decode is True, the raw AIS messages
        will be decoded and the decoded messages
        will be returned as a DataFrame.
        
        """
        dyn_options = dict(
            sep=",",
            usecols=self.dynamic_columns
        )
        stat_options = dict(
            sep=",",
            usecols=self.static_columns
        )
        for dyn_path, stat_path in zip(self.sdyn,self.sstat):
            
            # Count rows of both files
            with open(dyn_path) as f, open(stat_path) as g:
                dyn_rows = sum(1 for _ in f)
                stat_rows = sum(1 for _ in g)
             
            chunksize_stat = int(stat_rows * DataLoader.chunkfrac) + 1
            chunksize_dyn = int(dyn_rows * DataLoader.chunkfrac) + 1
            dyn_options["chunksize"] = chunksize_dyn
            stat_options["chunksize"] = chunksize_stat
            
            if decode:
                # Decode dynamic data
                yield self.from_raw(dyn_path, stat_path)
            else:
                # Create a generator of pandas DataFrames
                dyniter = pd.read_csv(dyn_path,**dyn_options)
                statiter = pd.read_csv(stat_path,**stat_options)
                
                for i, (dc,sc) in enumerate(zip(dyniter,statiter)):
                    logger.info(
                        f"Processing chunk {i+1} of file "
                        f"{dyn_path.stem}"
                    )
                    dc = self._dynamic_preprocessor(dc)            
                    yield dc, sc

# DEPRECATED v ================================================================
class Loader:
    def __init__(self, bb):
        """
        A loader-like context manager

        Args:
            desc (str, optional): The loader's description. Defaults to "Loading...".
            end (str, optional): Final print. Defaults to "Done!".
            timeout (float, optional): Sleep time between prints. Defaults to 0.1.
        """
        self.desc = (
            f"Buffering area from {bb.LATMIN:.3f}°N-{bb.LATMAX:.3f}°N "
            f"and {bb.LONMIN:.3f}°E-{bb.LONMAX:.3f}°E"
        )
        self.timeout = 0.1

        self._thread = Thread(target=self._animate, daemon=True)
        self.steps = ["⢿", "⣻", "⣽", "⣾", "⣷", "⣯", "⣟", "⡿"]
        self.done = False

    def start(self):
        self.t_start = time.perf_counter()
        self._thread.start()
        return self

    def _animate(self):
        for c in cycle(self.steps):
            if self.done:
                break
            print(f"{self.desc} {c}", flush=True, end="\r")
            time.sleep(self.timeout)

    def __enter__(self):
        self.start()

    def stop(self):
        self.done = True
        print(" "*100, end = "\r")
        logger.info(f"{self.desc}")
        self.t_end = time.perf_counter()
        logger.info(f"Cell Buffering completed in [{(self.t_end-self.t_start):.1f} s]")

    def __exit__(self, exc_type, exc_value, tb):
        # handle exceptions with those variables ^
        self.stop()

# Context manager that continuously shows memory usage
# while running the code inside the context.
class MemoryLoader:
    def __init__(self):
        self.timeout = 0.2

        self._thread = Thread(target=self._show_memory_usage, daemon=True)
        self.done = False

    def start(self):
        self.t_start = time.perf_counter()
        self._thread.start()
        return self

    def _show_memory_usage(self):
        """
        Prints memory usage every `timeout` seconds
        """
        while True:
            if self.done:
                break
            print(
                f"Memory usage: {psutil.virtual_memory().percent}% "
                f"[{psutil.virtual_memory().used/1e9:.2f} GB]", 
                end="\r")
            time.sleep(self.timeout)

    def __enter__(self):
        self.start()

    def stop(self):
        self.done = True
        print(" "*100, end = "\r")
        self.t_end = time.perf_counter()
        print(
            f"Loading took {self.t_end - self.t_start:.2f} seconds \n"
            f"Memory usage: {psutil.virtual_memory().percent}% "
            f"[{psutil.virtual_memory().used/1e9:.2f} GB]"
            )

    def __exit__(self, exc_type, exc_value, tb):
        # handle exceptions with those variables ^
        self.stop()


"""
Plotting functions for Empirical Cumulative 
Distribution Functions (ECDFs) of the split
point metrics.

All functions in this module are take an instance
of the :class:`pytsa.tsea.search_agent.SearchAgent`
as their only argument.
"""
import sys
from typing import Generator
import numpy as np
import matplotlib.patches as patches
from functools import partial
if sys.version_info >= (3,10):
    from itertools import pairwise
else:
    from more_itertools import pairwise

from . import (
    PLOT_FOLDER,
    COLORWHEEL,
    plt
)
from ..tsea.search_agent import SearchAgent, Targets
from ..tsea.targetship import TargetShip
from ..structs import AISMessage, BoundingBox
from ..tsea import split

def iter_msg(sa: SearchAgent, 
             skip_tsplit: bool = False
            ) -> Generator[tuple[AISMessage,AISMessage],None,None]:
    """
    Iterate over all messages in the 
    trajectories extacted by the `SearchAgent`
    and yield the raw messages in a pairwise
    fashion.
    """
    ships: Targets = sa.extract_all(skip_tsplit=skip_tsplit)
    for ship in ships.values():
        for track in ship.tracks:
            for msg1, msg2 in pairwise(track):
                yield (msg1,msg2)

# Export raw and split messages iterators         
iter_msg_raw = partial(iter_msg,skip_tsplit=True)
iter_msg_tsplit = partial(iter_msg,skip_tsplit=False)

def plot_time_diffs(sa: SearchAgent):
    """
    Plot the time difference between two consecutive
    messages for data in the `SearchAgent`.
    """
    f, ax = plt.subplots(1,1,figsize=(6,4))
    ax: plt.Axes
    time_diffs = []
                
    for msg1, msg2 in iter_msg_raw(sa):
        time_diffs.append(msg2.timestamp - msg1.timestamp)
        
    # Quantiles of time diffs
    qs = np.quantile(
        time_diffs,
        [0.99,0.95,0.90]
    )

    # Boxplot of time diffs
    ax.boxplot(
        time_diffs,
        vert=False,
        showfliers=False,
        patch_artist=True,
        widths=0.5,
        boxprops=dict(facecolor=COLORWHEEL[0], color=COLORWHEEL[0]),
        medianprops=dict(color=COLORWHEEL[1]),
        whiskerprops=dict(color=COLORWHEEL[1]),
        capprops=dict(color=COLORWHEEL[1]),
        flierprops=dict(color=COLORWHEEL[1], markeredgecolor=COLORWHEEL[1])
    )
    # Remove y ticks
    ax.set_yticks([])
    
    # Legend with heading
    ax.legend(
        handles=[
            patches.Patch(
                color=COLORWHEEL[0],
                label=f"1% larger than {qs[0]:.2f} s"
            ),
            patches.Patch(
                color=COLORWHEEL[0],
                label=f"5% larger than {qs[1]:.2f} s"
            ),
            patches.Patch(
                color=COLORWHEEL[0],
                label=f"10% larger than {qs[2]:.2f} s"
            )
        ]
    )
    
    ax.set_xlabel("Time difference [s]")
    
    ax.set_title(
        "Time difference between two consecutive messages",fontsize=10
    )
    
    plt.tight_layout()
    plt.savefig(f"{PLOT_FOLDER}/time_diffs.pdf")
    
def plot_reported_vs_calculated_speed(sa:SearchAgent):
    """
    Plot the reported speed against the
    calculated speed.
    """
    
    f, ax = plt.subplots(1,1,figsize=(6,4))
    ax: plt.Axes
    speeds = []
    
    for msg1, msg2 in iter_msg_raw(sa):
        rspeed = split.avg_speed(msg1,msg2)
        cspeed = split.speed_from_position(msg1,msg2)
        speeds.append(rspeed - cspeed)
    
    # Boxplot of speeds
    ax.boxplot(
        speeds,
        vert=False,
        labels=["Speed"],
        showfliers=False,
        patch_artist=True,
        widths=0.5,
        boxprops=dict(facecolor=COLORWHEEL[0], color=COLORWHEEL[0]),
        medianprops=dict(color=COLORWHEEL[1]),
        whiskerprops=dict(color=COLORWHEEL[1],width=1.5),
        capprops=dict(color=COLORWHEEL[1]),
        flierprops=dict(color=COLORWHEEL[1], markeredgecolor=COLORWHEEL[1])
    )
    
    # Quantiles of speeds
    s_qs = np.quantile(
        speeds,
        [0.005,0.995,0.025,0.975,0.05,0.95]
    )
    q_labels_h = [
        f"99% within [{s_qs[0]:.2f} kn,{s_qs[1]:.2f} kn]",
        f"95% within [{s_qs[2]:.2f} kn,{s_qs[3]:.2f} kn]",
        f"90% within [{s_qs[4]:.2f} kn,{s_qs[5]:.2f} kn]"
    ]

    # Legend with heading
    ax.legend(handles=[
        patches.Patch(color=COLORWHEEL[0],label=q_labels_h[0]),
        patches.Patch(color=COLORWHEEL[0],label=q_labels_h[1]),
        patches.Patch(color=COLORWHEEL[0],label=q_labels_h[2])
    ])
    
    ax.set_xlabel("Difference [kn]")
    
    ax.set_title(
        "Difference between reported and calculated speed [kn]",fontsize=10
    )
    
    plt.tight_layout()
    plt.savefig(f"{PLOT_FOLDER}/reported_vs_calculated_speed.pdf")
    
def plot_heading_and_speed_changes(sa: SearchAgent):
    """
    Plot the changes in heading and speed between
    two consecutive messages.
    
    For historical reasons, this function does two
    things at once.
    Feel free to split this function into two separate.
    """
    
    def _heading_change(h1,h2):
        """
        calculate the change between two headings
        such that the smallest angle is returned.
        """
        diff = abs(h1-h2)
        if diff > 180:
            diff = 360 - diff
        if (h1 + diff) % 360 == h2:
            return diff
        else:
            return -diff
    f, ax = plt.subplots(1,1,figsize=(8,5))

    heading_changes = []
    speed_changes = []
                
    for msg1, msg2 in iter_msg_raw(sa):
        _chheading = _heading_change(
            msg1.COG,
            msg2.COG
        )
        _chspeed = abs(msg2.SOG - msg1.SOG)
        heading_changes.append(_chheading)
        speed_changes.append(_chspeed)

    # Quantiles of heading changes
    h_qs = np.quantile(
        heading_changes,
        [0.005,0.995,0.025,0.975,0.05,0.95]
    )
    q_labels_h = [
        f"99% within [{h_qs[0]:.2f}°,{h_qs[1]:.2f}°]",
        f"95% within [{h_qs[2]:.2f}°,{h_qs[3]:.2f}°]",
        f"90% within [{h_qs[4]:.2f}°,{h_qs[5]:.2f}°]"
    ]
    # Heading Quantiles as vertical lines
    hl11 = ax.axvline(h_qs[0],color=COLORWHEEL[0],label=q_labels_h[0],ls="--")
    hl21 = ax.axvline(h_qs[2],color=COLORWHEEL[0],label=q_labels_h[1],ls="-.")
    hl31 = ax.axvline(h_qs[4],color=COLORWHEEL[0],label=q_labels_h[2],ls=":")
    
    # Histogram of heading changes
    ax.hist(
        heading_changes,
        bins=100,
        density=True,
        alpha=0.8,
        color=COLORWHEEL[0]
    )
    
    # Quantiles of speed changes
    s_qs = np.quantile(
        speed_changes,
        [0.99,0.95,0.90]
    )
    
    q_labels_s = [
        f"99% smaller than {s_qs[0]:.2f} kn",
        f"95% smaller than {s_qs[1]:.2f} kn",
        f"90% smaller than {s_qs[2]:.2f} kn"
    ]
    
    
    # Legend with heading
    ax.legend(handles=[hl11,hl21,hl31])
    ax.set_xlabel("Change in heading [°]")
    ax.set_ylabel("Density")
    ax.set_title(
        "Change in heading between two consecutive messages",fontsize=10
    )
    plt.tight_layout()
    plt.savefig(f"{PLOT_FOLDER}/heading_changes.pdf")
    plt.close()
    
    # Plot speed changes
    f, ax = plt.subplots(1,1,figsize=(8,5))

    # Speed Quantiles as vertical lines
    sl1 = ax.axvline(s_qs[0],color=COLORWHEEL[0],label=q_labels_s[0],ls="--")
    sl2 = ax.axvline(s_qs[1],color=COLORWHEEL[0],label=q_labels_s[1],ls="-.")
    sl3 = ax.axvline(s_qs[2],color=COLORWHEEL[0],label=q_labels_s[2],ls=":")
    
    # Histogram of speed changes
    ax.hist(
        speed_changes,
        bins=200,
        density=True,
        alpha=0.8,
        color=COLORWHEEL[0]
    )
    
    ax.legend(handles=[sl1,sl2,sl3])
    ax.set_xlabel("Absolute change in speed [knots]")
    ax.set_ylabel("Density")
    ax.set_title(
        "Change in speed between two consecutive messages",fontsize=10
    )
    ax.set_xlim(-0.2,4)
    
    plt.tight_layout()
    plt.savefig(f"{PLOT_FOLDER}/speed_changes.pdf")
    



"""
Target ship extraction agent (:mod:`pytsa.tsea`)
=============================

This module contains provides twofold functionality:
    1. Extract target ship trajectories from raw AIS data
    2. Query for target ships around a given position and time

Both functionalities rely on the :class:`SearchAgent` class, 
which is initialized with the AIS data and a geographic bounding box.

AIS DATA ------------------------------
The AIS data are expected to be decoded CSV files, split up into two files:
    1. Dynamic data (messages 1,2,3,18)
    2. Static data (message 5)
    
If your messages are not decoded yet, please refer to the
:mod:`pytsa.decode` module.
"""


"""
This module defines some structures and functions 
to decode, sort, and export AIS Messages.

The structures are specific to the EMSA dataset
being analyzed with it. 
"""
import os
from typing import Any, List, Tuple, Dict
from pathlib import Path

from enum import Enum
import numpy as np
import pandas as pd
import pyais as ais
import pyais.messages as msg
import multiprocessing as mp
from pytsa.decode.filedescriptor import (
    BaseColumns
)

class StructuralError(Exception):
    pass

# Message type tuples
_STATIC_TYPES = (5,)
_DYNAMIC_TYPES = (1,2,3,18,)

# Slot names for message fields
MSG5SLOTS = msg.MessageType5.__slots__
MSG12318SLOTS = msg.MessageType1.__slots__
MSGSLOTS = tuple[str]

# Default value for missing data
_NA = "NA"

# Decoding plan
class DynamicDecoder:
    type = "dynamic"
    
    def __call__(self, df: pd.DataFrame) -> Any:
        return self._decode_dynamic_messages(df)

    def _decode_dynamic_messages(self, df: pd.DataFrame) -> List[ais.ANY_MESSAGE]:
        """
        Decode AIS messages of types 1,2,3,18 
        supplied as a pandas Series object.
        """
        messages = df[BaseColumns.RAW_MESSAGE.value]
        # Split at exclamation mark and take the last part
        raw = messages.str.split("!",expand=True)
        raw = raw.dropna(axis=1).iloc[:,-1] # Catch NaNs
        # Since we split on the exclamation mark we need to
        # re-add it at the front of the message
        raw = "!" + raw
        
        out = []
        to_drop = []
        for idx,val in enumerate(raw.values.ravel()):
            try:
                out.append(ais.decode(val))
            except Exception as e:
                print(f"Error decoding {val}: {e}")
                to_drop.append(idx)
        
        return out, to_drop

class StaticDecoder:
    type = "static"
    
    def __call__(self, df: pd.DataFrame) -> Any:
        return self._decode_static_messages(df)
    
    def _decode_static_messages(self, df: pd.DataFrame):
        """
        Decode AIS messages of type 5 
        supplied as a pandas DataFrame.
        """
        msg1 = df[BaseColumns.RAW_MESSAGE1.value]
        msg2 = df[BaseColumns.RAW_MESSAGE2.value]
        raw1 = msg1.str.split("!",expand=True).iloc[:,-1:]
        raw2 = msg2.str.split("!",expand=True).iloc[:,-1:]
        raw1, raw2 = "!" + raw1, "!" + raw2
        
        out = []
        to_drop = []
        for idx, vals in enumerate(zip(raw1.values.ravel(),raw2.values.ravel())):
            try:
                out.append(ais.decode(*vals))
            except Exception as e:
                print(f"Error decoding {vals}: {e}")
                to_drop.append(idx)
        
        return out, to_drop

# Type alias
Decoder = DynamicDecoder | StaticDecoder

def _extract_fields(messages: List[ais.ANY_MESSAGE],
                    fields: tuple) -> Dict[str,np.ndarray]:
    """
    Get the values of the fields in the messages
    """
    out = np.empty((len(messages),len(fields)),dtype=object)
    for i, msg in enumerate(messages):
        line = []
        for field in fields:
            attr = getattr(msg,field,_NA)
            # The pyais library uses the enum module
            # to define some of the fields. We need to
            # unpack them to get the actual value.
            if isinstance(attr,Enum): # unpack enum
                attr = attr.value
            line.append(attr)
        out[i] = line
    return dict(zip(fields,out.T))

def _get_decoder(dataframe: pd.DataFrame) -> Tuple[Decoder,MSGSLOTS]:
    """
    Returns a message-specific decoding function
    based on message types present in the dataframe.

    Note: Since input dataframes will be processed 
    at once, they can only contain either position 
    record messages (types 1,2,3,18), or static messages
    (type 5) but not both.
    """
    types = dataframe[BaseColumns.MESSAGE_ID.value]
    if all(k in dataframe for k in (BaseColumns.RAW_MESSAGE1.value, 
        BaseColumns.RAW_MESSAGE2.value)):
        # Maybe type 5 
        if all(b in _STATIC_TYPES for b in types.unique()):
            return StaticDecoder(), MSG5SLOTS
        else: raise StructuralError(
                "Assumed type-5-only dataframe, but found "
                f"messages of types {types.unique()}"
        )
    elif all(b in _DYNAMIC_TYPES for b in types.unique()):
        return DynamicDecoder(), MSG12318SLOTS
    else: raise StructuralError(
            "Found not processable combination "
            "of message types. Need either type-5-only dataframe "
            "or type-1-2-3-18-only dataframe. Found messages "
            f"of types {types.unique()}"
    )
    
# Pandas file operations 
#
# The pipeline will first read-in the csv-file
# as a pandas dataframe, decode the raw AIS
# message, and save the extracted information 
# as new columns in the existing file. 
def decode_from_file(source: str,
                     dest: str,
                     save_to_file = True) -> None:
    print(f"Decoding {source} to {dest}")
    df  = pd.read_csv(source,sep=",",quotechar='"',
                      encoding="utf-8",index_col=False)
    # There are no newline characters allowed in the csv
    df = df.replace(r'\n','', regex=True)
    
    dropnasubset = [BaseColumns.RAW_MESSAGE1.value,
                    BaseColumns.RAW_MESSAGE2.value] if \
                    BaseColumns.RAW_MESSAGE1.value in df.columns else \
                    [BaseColumns.RAW_MESSAGE.value]
    df = df.dropna(subset=dropnasubset)
    decoder, fields = _get_decoder(df)
    decoded, to_drop = decoder(df)
    df = df.drop(index=to_drop) # Drop messages that could not be decoded
    df["DECODE_START"] = "||"
    df = df.assign(**_extract_fields(decoded,fields))
    if save_to_file:
        df.to_csv(dest,index=False)
        del df
        return
    else:
        return df

def decode(source: Path, 
           dest: Path, njobs: int = 16, 
           overwrite: bool = False) -> None:
    """
    Decode AIS messages in parallel.

    overwrite: If True, overwrite existing files in the destination folder.
    """
    files = list(source.glob("*.csv")) 
    # Check if any files are in the destination folder
    # and remove them from the list of files to be processed
    # to avoid overwriting data.
    if overwrite:
        if dest.exists():
            dnames = [f.name for f in list(dest.glob("*.csv"))]
            files = [f for f in files if f.name not in dnames]
    
    if njobs == 1:
        for file in files:
            decode_from_file(
                file, 
                f"{dest.as_posix()}/{'/'.join(file.parts[len(source.parts):])}"
            )
        return
    else: 
        with mp.Pool(processes=njobs) as pool:
            pool.starmap(
                decode_from_file, 
                [(file, f"{dest.as_posix()}/{'/'.join(file.parts[len(source.parts):])}")
                        for file in files]
            )
        return

if __name__ == "__main__":
    SOURCE = Path(os.environ["AISSOURCE"])
    DEST = Path(os.environ["DECODEDDEST"])
    decode(SOURCE,DEST,njobs=32)